--########################################################################
--
--  PROCESSADOR CLEPATRA
--
--  DISPONVEL a partir de:  http://www.inf.pucrs.br/~calazans
--
--  FERNANDO GEHM MORAES   03/agosto/2005
--  NEY CALAZANS			15/abril/2006
--
--########################################################################



-------------------------------------------------------------------------
-- Registrador Bsico
-- Com reset assncrono e habilitao de escrita
-------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity reg_clear is
    port(clock, reset, ce:in std_logic; D:in internal_bus; Q:out internal_bus );
end reg_clear;

architecture reg_clear of reg_clear is
begin
  process (clock, reset)
    begin
    if reset='1' then
         Q <= (others=>'0');
    elsif (clock'event and clock='0')then
        if ce='1' then  Q <= D;  end if;
    end if;
  end process;
end reg_clear;


-------------------------------------------------------------------------
-- Decodificador de Leitura
-- Este bloco combinacional gera sinais de habilitao de dispositivos
-- tristate interpostos entre os registradores e barramentos.
-------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity read_decoder is
    port(read_reg:in std_logic_vector(2 downto 0);
        r_mdr, r_ir, r_pc, r_ac, r_rs: out std_logic);
end read_decoder;

architecture read_decoder of read_decoder is
begin
    ---
    --- Cdigos de leitura nos registradores (codificao de leitura):
    --- 1 MDR, 2 IR, 3 PC, 4 AC, 5 RS, 6 MDR/AC, 7 MDR/PC
    ---
    r_mdr <= '1' when read_reg="001" or read_reg="110" or read_reg="111" else '0';
    r_ir  <= '1' when read_reg="010"                                     else '0'; 
    r_pc  <= '1' when read_reg="011" or read_reg="111"                   else '0';
    r_ac  <= '1' when read_reg="100" or read_reg="110"                   else '0';  
    r_rs  <= '1' when read_reg="101"                                     else '0';
end read_decoder;                                                                                                                                        

-------------------------------------------------------------------------
-- Decodificador de Escrita
-- Este bloco combinacional gera sinais de habilitao de escrita nos 
-- registradores do Bloco de Dados.
--
--  30/07/2004 - reviso para considerar o sinal de hold na interface
--              externa do processador. (Calazans)
-------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

--  30/07/2004 - a escrita de regsitradores deve ser condicionada
--               inexistncia de ativao de hold. (Calazans)
--      CUIDADO: O sinal hold somente deve ser ativado pelo mundo externo
--              durante um ciclo de leitura da memria (ce='1', rw='1').
--              Se isto no for respeitado, funcionamento incorreto do
--              processador pode ocorrer devido ao tratamento feito neste mdulo.
entity write_decoder is
    port(write_reg :in std_logic_vector(2 downto 0); hold :in std_logic;
        w_mar, w_mdr, w_ir, w_pc, w_ac, w_rs :out std_logic);
end write_decoder;

architecture write_decoder of write_decoder is
begin
    ---
    --- Cdigos de escrita nos registradores (codificao de escrita):
    --- 0 MAR, 1 MDR, 2 IR, 3 PC, 4 AC, 5 RS, 6 PC/MDR, 7 nenhum reg. alterado
    ---
    w_mar <= '1' when hold='0' and write_reg="000"                          else '0';
    w_mdr <= '1' when hold='0' and (write_reg="001" or write_reg="110") else '0';
    w_ir  <= '1' when hold='0' and write_reg="010"                              else '0';
    w_pc  <= '1' when hold='0' and (write_reg="011" or write_reg="110")         else '0';
    w_ac  <= '1' when hold='0' and write_reg="100"                              else '0';
    w_rs  <= '1' when hold='0' and write_reg="101"                              else '0';    
end write_decoder;                                                                                                                                       


-------------------------------------------------------------------------
-- Unidade lgico-aritmtica (ALU)
-- A ALU possui duas entradas de dados (A_bus e B_bus) e 
-- uma saida de dados (C_bus), um sinal de controle de operao a executar
-- (alu_op) e quatro sadas com valores instantneos computados para os
-- 4 qualificadores de estado (n_inst, z_inst, c_inst, v_inst).
-------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use work.cleo.all;

entity alu is
    port(A_bus, B_bus:in internal_bus; alu_op: in std_logic_vector(2 downto 0);
        C_Bus:out internal_bus; n_inst, z_inst, c_inst, v_inst: out std_logic);
end alu;

architecture alu of alu is
-- barramentos internos possuem um bit a mais para permitir clculo dos flags
-- de vai-um e transbordo
signal A_bus_int, B_bus_int, C_bus_int : std_logic_vector(bus_size downto 0);

begin
     
     A_bus_int <= '0' & A_bus; -- zera-se o bit mais significativo dos
     B_bus_int <= '0' & B_bus; -- barramentos internos de entrada
     C_bus <= C_bus_int (bus_size-1 downto 0); -- toma n-1 bits menos significativos

     process(alu_op, A_bus_int, B_bus_int)
     begin
       case alu_op is
        when "000" => C_bus_int <= A_bus_int + B_bus_int; -- soma
        when "001" => C_bus_int <= A_bus_int + "000000001"; -- incremento
        when "010" => C_bus_int <= not A_bus_int; -- not
        when "100" => C_bus_int <= B_bus_int;      -- passa B
        when "101" => C_bus_int <= A_bus_int or B_bus_int;      -- ou lgico
        when "110" => C_bus_int <= A_bus_int and B_bus_int;     -- e lgico
        when "111" => C_bus_int <= A_bus_int;      -- passa A
        when others => C_bus_int <= A_bus_int;     -- faz passa A;
       end case;
     end process;                                         
    
    -- o bit menos significativo-1  o valor instantneo do qualificador n
    n_inst <= C_bus_int (bus_size-1);
    
    -- o qualificador z em 1 indica que o resultado da operao  o valor 0
    z_inst <= '1'   when (C_bus_int(bus_size-1 downto 0) = "00000000") else 
              '0';
    -- o bit menos significativo  o valor instantneo do qualificador c
    c_inst <= C_bus_int (bus_size);
    
    -- o qualificador v vai para 1 se os operandos tm sinais iguais 
    -- e se este sinal  diferente do sinal do resultado
     v_inst <= '1' when (A_bus_int(bus_size-1)=B_bus_int(bus_size-1) and 
                        A_bus_int(bus_size-1)/=C_bus_int(bus_size-1)) else 
               '0';    
end alu;

-------------------------------------------------------------------------
-- Biestveis de estado
-- Este registrador de estado armazena quatro bits referentes aos 4
-- valores instantneos dos qualificadores de estado
-- (n_inst, z_inst, c_inst, v_inst) quando os sinais de controle 
-- (lnz e lcv) indicam que isto deva ocorrer, na borda correta do sinal de
-- relgio (clock). O sinal reset inicializa todos os qualificadores com 0.
-------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity status_flags is
    port(n_inst, z_inst, c_inst, v_inst, lnz, lcv, clock, reset:in std_logic;
        n, z, c, v: out std_logic);
end status_flags;

architecture status_flags of status_flags is
begin
    process(clock,reset)
    begin
        if (reset='1') then
          n <= '0';
          z <= '0';
          c <= '0';    
          v <= '0';    
        elsif clock'event and clock='0' then
          if lcv='1' then
              c <= c_inst;
              v <= v_inst;
          end if;
          if lnz='1' then
             n <= n_inst;
             z <= z_inst;
          end if;
        end if;
    end process;                   
end status_flags;

--------------------------------------------------------------------------
-- Bloco de Dados
-- Descricao estrutural do bloco de dados do processador Clepatra,
-- pela interconexo de vrias de suas partes e criao do Mux que
-- gera entrada para o MDR e dos trs barramentos internos
--------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity datapath is
    port(clock, reset, hold     : in std_logic;
         address, ir    : out internal_bus;
         datain                 : in internal_bus;
         dataout                : out internal_bus;
         uins                   : in microinstrucao;
         n, z, c, v             : out std_logic
        ); 
end datapath;

architecture datapath of datapath is
signal  r_mdr,r_ir,r_pc,r_ac,r_rs,
                w_mar,w_mdr,w_ir,w_pc,w_ac,w_rs,
                n_inst, z_inst, c_inst, v_inst : std_logic; 
signal  outmux, mdr, ir_int, pc, ac, rs,
                busA, busB, busC : internal_bus;
 
begin  

    ir <= ir_int;    -- instrucao corrente, a ser utilizada no bloco de controle
        
    --
    -- barramento de sada do processador  sempre a sada a ALU
    --
    dataout <= busC;
    
    --
    -- cria multiplexador que gera entrada para o registrador MDR e seu controle
    --
    outmux <= busC  when (uins.ce and uins.rw)='0' else datain;
                
    --
    -- cria registradores de dados (mar, mdr, ir, pc, ac, rs),
    -- instancia-os e conecta-os 
    --
    REG_MAR: entity work.reg_clear port map (clock=>clock, reset=>reset, ce=>w_mar, D=>busC, Q=>address);
    REG_MDR: entity work.reg_clear port map (clock=>clock, reset=>reset, ce=>w_mdr, D=>outmux,  Q=>mdr);
    REG_IR:  entity work.reg_clear port map (clock=>clock, reset=>reset, ce=>w_ir,  D=>busC, Q=>ir_int);
    REG_PC:  entity work.reg_clear port map (clock=>clock, reset=>reset, ce=>w_pc,  D=>busC, Q=>pc);
    REG_AC:  entity work.reg_clear port map (clock=>clock, reset=>reset, ce=>w_ac,  D=>busC, Q=>ac);
    REG_RS:  entity work.reg_clear port map (clock=>clock, reset=>reset, ce=>w_rs,  D=>busC, Q=>rs);
    
    --
    -- instancia decodificador de escrita e conecta com habilitaes de
    -- escrita dos registradores
    --
    W_DECOD: entity work.write_decoder port map(write_reg=>uins.w, 
                hold=>hold,
                w_mar=>w_mar, w_mdr=>w_mdr, w_ir=>w_ir,
                w_pc=>w_pc, w_ac=>w_ac, w_rs=>w_rs);
    
    --
    -- instancia decodificador de leitura e conecta com habilitaes de
    -- tristates que controlam o que vai para barramentos busA e busB
    --
    R_DECOD: entity work.read_decoder port map(read_reg=>uins.r,
                r_mdr=>r_mdr, r_ir=>r_ir, r_pc=>r_pc, r_ac=>r_ac, r_rs=>r_rs);
                
    --
    -- cria tristates para jogar o valor de 1 ou 2 registradores
    -- nos barramentos de entrada da ALU, busA e busB
    --
    busB <= mdr     when r_mdr='1' else (others=>'Z');
    busB <= ir_int  when r_ir='1'  else (others=>'Z');
    
    busA <= pc      when r_pc='1'  else (others=>'Z');
    busA <= ac      when r_ac='1'  else (others=>'Z');
    busA <= rs      when r_rs='1'  else (others=>'Z');
    
    --
    -- instancia ALU e a conecta com barramentos busA, busB e busC,
    -- bem como com a entrada de controle e o registrador de estado
    --
    ALU: entity work.alu port map(A_bus=>busA, B_bus=>busB, C_bus=>busC,
                alu_op=>uins.u, n_inst=>n_inst, z_inst=>z_inst,
                c_inst=>c_inst, v_inst=>v_inst);

    --
    -- instancia o registrador de estado e conecta-o  ALU e a pinos da 
    -- interface com o bloco de controle
    --
    REG_STATUS: entity work.status_flags port map(n_inst=>n_inst, 
                z_inst=>z_inst, c_inst=>c_inst, v_inst=>v_inst, lnz=>uins.lnz,
                lcv=>uins.lcv, clock=>clock, reset=>reset, n=>n, z=>z,
                c=>c, v=>v);

end datapath;


--------------------------------------------------------------------------
-- Bloco de Controle do Processador Clepatra (8 bits)
-- Soluo hardwired, via mquina de estados 
--
--  1/08/2004 - reviso para sntese em hardware. trocar processo nico
--  com muitos waits por estrutura de FSM convencional. (Calazans)
--
--  14/8 ==> mquina de estados simplificada (15 estados) (moraes)
--  15/8 ==> insero do sinal hold e codificao extendida para not/sta
--------------------------------------------------------------------------
library IEEE; 
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity control is
    port( reset, clock, hold : in std_logic;
          halt: out std_logic;
          ir :   in internal_bus;
          n, z, c, v: in std_logic;
          uins : out microinstrucao
        ); 
end control;

architecture control of control is 

   type CleoStates_type is (
       SIDLE,                             -- estado dummy, apenas para garantir que borda de subida seguinte
                                          --  desativao do reset  o que faz o processamento comear.
       FETCH0, FETCH1, FETCH2,            -- estados onde se faz busca de instruo
       
       Sop1a, Sop1b,                      --  busca do primeiro operando das demais instrues
       
       Sop2a, Sop2b, Sop3a, Sop3b, Sula,  -- estados para execuo das instrues lgicas aritmticas 
          
       OP2jp1, OP2jp2, Ssalta,            -- estados para execuo dos saltos
       
       Sjsr
   );
   
   -- CODIFICAO DAS INSTRUCOES
   constant NOT1  : std_logic_vector(3 downto 0) :=  x"0";     -- duas codificaes possveis para o NOT
   constant NOT2  : std_logic_vector(3 downto 0) :=  x"1";   
   constant STA1  : std_logic_vector(3 downto 0) :=  x"2";     -- duas codificaes possveis para o STA
   constant STA2  : std_logic_vector(3 downto 0) :=  x"3";  
   constant LDA   : std_logic_vector(3 downto 0) :=  x"4";  
   constant ADD   : std_logic_vector(3 downto 0) :=  x"5";
   constant ORL   : std_logic_vector(3 downto 0) :=  x"6";
   constant ANDL  : std_logic_vector(3 downto 0) :=  x"7";
   constant JMP   : std_logic_vector(3 downto 0) :=  x"8";     
   constant JC    : std_logic_vector(3 downto 0) :=  x"9";     
   constant JN    : std_logic_vector(3 downto 0) :=  x"A";
   constant JZ    : std_logic_vector(3 downto 0) :=  x"B";
   constant JSR   : std_logic_vector(3 downto 0) :=  x"C";
   constant RTS   : std_logic_vector(3 downto 0) :=  x"D";
   constant JV    : std_logic_vector(3 downto 0) :=  x"E";
   constant HLT   : std_logic_vector(3 downto 0) :=  x"F";

   -- CODIFICAO DOS MODOS DE ENDEREAMENTO
   constant IM    : std_logic_vector(1 downto 0) :=  "00";
   constant DIR   : std_logic_vector(1 downto 0) :=  "01";    
   constant IND   : std_logic_vector(1 downto 0) :=  "10";
   constant REL   : std_logic_vector(1 downto 0) :=  "11";    

   --
   -- definicao das microinstrucoes em funcao dos registradores destino/origem
   --                                       wr     rd     alu  ce  rw  lnz lcv
   constant mar_pc    : microinstrucao := ("000", "011", "111",'0','0','0','0');  -- mar <- pc  
   constant mar_mdr   : microinstrucao := ("000", "001", "100",'0','0','0','0');  -- mar <- mdr
   constant mdr_MmarP : microinstrucao := ("110", "011", "001",'1','1','0','0');  -- mdr <- pmem(mar); pc++
   constant mdr_Mmar  : microinstrucao := ("001", "110", "111",'1','1','0','0');  -- mdr <- pmem(mar)
   constant mar_mdrpc : microinstrucao := ("000", "111", "000",'0','0','0','0');  -- mar <- mdr+PC 
   constant ir_mdr    : microinstrucao := ("010", "001", "100",'0','0','0','0');  -- ir  <- mdr
   constant pc_mdr    : microinstrucao := ("011", "001", "100",'0','0','0','0');  -- pc  <- mdr
   constant pc_mdrpc  : microinstrucao := ("011", "111", "000",'0','0','0','0');  -- pc  <- mdr + pc
   constant rts_pc    : microinstrucao := ("101", "011", "111",'0','0','0','0');  -- rts <- pc
   constant nop       : microinstrucao := ("111", "110", "111",'0','0','0','0');  -- nao faz nada 

   alias i  : std_logic_vector(3 downto 0) is ir(7 downto 4);     -- pedaos do IR
   alias me : std_logic_vector(1 downto 0) is ir(3 downto 2);

   signal salta, logic_arith : std_logic;                         -- auxiliares

   signal EA, PE: CleoStates_type;                                -- estados da mquina de estado

begin
       
   salta <= '1'  when ((i=JC and c='1') or (i=JV and v='1') or (i=JN and n='1') or (i=JZ and z='1')) else
            '0';
             
   logic_arith <= '1' when  i=LDA or i=ADD or i=ORL or i=ANDL else
                  '0';
   
   -- processo para implementar registrador de estado 
   process (clock, reset)
    begin
         if reset='1' then
              EA  <= Sidle; -- Sidle is the state the machine stays while processor is being reset
              halt <= '0';
         elsif clock'event and clock='1' then
              if i=HLT then 
                  halt <= '1';        -- avisa que trancou
              elsif hold='0' then     
                  EA <= PE;           -- O PROCESSADOR TRANCA EM HOLD!!
              end if;  
         end if;
   end process;
  
  -- processo para implementar funes de sada e prximo estado
  process (hold, i, salta, EA)
  begin                  
     case EA is
          
       when SIDLE  => uins <= nop; 
                      PE <= FETCH0;                       -- no faz nada em SIDLE

       --- 
       --- BUSCA DA INSTRUO
       ---             
       when FETCH0 => uins <= mar_pc;                     -- MAR <- PC              primeiro ciclo da busca
                      PE <= FETCH1;
                       
       when FETCH1 => uins <= mdr_MmarP;                  -- MDR <- MEM(MAR); PC++  segundo ciclo da busca.
                      PE <= FETCH2;
              
       when FETCH2 => uins <= ir_mdr;                     -- IR <- MDR              carga do ir
                  
                      case i is
                            when RTS | NOT1 | NOT2 | HLT =>  PE <=  Sula;
                            when others                  =>  PE <=  Sop1a;
                      end case;                                          
       --- 
       --- BUSCA DO OPERANDO DA INSTRUO
       ---  
       when Sop1a =>  uins <= mar_pc;        -- MAR <- PC 
                      PE <=  Sop1b;
        
       when Sop1b =>  uins <= mdr_MmarP;     -- MDR <- MEM(MAR); PC++
       
                      if logic_arith='1' and ME=IM then
                             PE <= Sula;
                      elsif  (logic_arith='1' and (ME=DIR or ME=IND or ME=REL)) or (i=STA1 or i=STA2) then
                             PE <= Sop2a;
                      elsif  (i=JMP  or salta='1') and (ME=IM or ME=DIR or ME=REL) then
                             PE <= Ssalta;
                      elsif  (i=JMP  or salta='1') and  ME=IND then
                             PE <= OP2jp1;
                      elsif i=JSR then
                             PE <= Sjsr;
                      else
                             PE <= FETCH0;    -- volta para o FETCH, e.g., salto condicional com flag=0    
                      end if;    
      --- 
      --- AES ESPECFICAS DAS INSTRUES LGICAS ARITMTICAS -- lda, and, or, add, not + STA + RTS
      ---
      when Sop2a =>   case ME is
                            when REL    =>  uins <= mar_mdrpc;    -- MAR <- MDR + PC
                            when others =>  uins <= mar_mdr;      -- MAR <- MDR
                      end case;                                          
 
                      if (i=STA1 or i=STA2) and (ME=REL or ME=DIR) then
                                PE <= Sula;
                        else
                                PE <=  Sop2b;
                      end if;
                        
      when Sop2b =>   uins <= mdr_Mmar;            -- MDR <- MEM(MAR)
                        
                      case ME is
                            when IND    =>  PE <=  Sop3a;
                            when others =>  PE <=  Sula;
                      end case;                                          
                                            
      when Sop3a =>   uins <= mar_mdr;             -- MAR <- MDR

                      case i is
                            when STA1 | STA2 =>  PE <=  Sula;
                            when others      =>  PE <=  Sop3b;
                      end case;                                          
           
      when Sop3b =>   uins <= mdr_Mmar;            -- MDR <- MEM(MAR)
                      PE <=  Sula;
         
      when Sula  =>   case i is 
                       when LDA         => uins <= ("100", "001", "100", '0','0', '1','0'); -- ac <- mdr; lnz
                       when ADD         => uins <= ("100", "110", "000", '0','0', '1','1'); -- ac <- ac + mdr;lnz;lcv
                       when ORL         => uins <= ("100", "110", "101", '0','0', '1','0'); -- ac <- ac or mdr; lnz
                       when ANDL        => uins <= ("100", "110", "110", '0','0', '1','0'); -- ac <- ac and mdr; lnz
                       when NOT1 | NOT2 => uins <= ("100", "100", "010", '0','0', '1','0'); -- ac <- not ac ; lnz
                       when RTS         => uins <= ("011", "101", "111", '0','0', '0','0'); -- pc <- rts
                       when STA1 | STA2 => uins <= ("111", "100", "111", '1','0', '0','0'); -- pmem(mar) <- ac
                       when others      => uins <= nop;                                   -- halt
                      end case;    
                      
                      PE <= FETCH0;
                    
      --- 
      --- AES ESPECFICAS DAS INSTRUO DE SALTO
      ---
      when Sjsr =>    uins <= rts_pc;              -- RS <- PC

                      case ME is
                            when IND    =>  PE <=  OP2jp1;
                            when others =>  PE <=  Ssalta;
                      end case;                                          
            
      when OP2jp1  => uins <= mar_mdr;             -- MAR <- MDR
                      PE <=  OP2jp2;
 
      when OP2jp2  => uins <= mdr_Mmar;            -- MDR <- MEM(MAR)
                      PE <=  Ssalta;
                        
      when Ssalta =>  case ME is
                            when REL    =>  uins <= pc_mdrpc;      -- PC <- MDR + PC
                            when others =>  uins <= pc_mdr;        -- PC <- MDR
                      end case; 
                                            
                      PE <= FETCH0;
     end case;
  end process; 
end control;               

--------------------------------------------------------------------------
-------------------------------------------------------------------------
--  CPU: uniao do bloco de dados com o bloco de controle
--------------------------------------------------------------------------
--------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity cleopatra is
        port( clock, reset, hold : in std_logic;
              halt                   : out std_logic;
              ce, rw                 : out std_logic; 
              address                : out internal_bus;
              datain                 : in internal_bus;
              dataout                : out internal_bus
             ); 
end cleopatra;
    
architecture cleopatra of cleopatra is
   signal uins : microinstrucao; 
   signal n,z,c,v : std_logic;
   signal ir : internal_bus;
   
begin
   ce <= uins.ce;
   rw <= uins.rw;

   DP: entity work.datapath port map (clock=>clock, reset=>reset,
           hold=>hold, address=>address, datain=>datain, dataout=>dataout,
           ir=>ir, uins=>uins, n=>n, z=>z, c=>c, v=>v);  
                           
   CU: entity work.control port map (clock=>clock, reset=>reset,
           hold=>hold, halt=>halt, ir=>ir, uins=>uins,
           n=>n, z=>z, c=>c, v=>v);
end cleopatra;
